"""Evolution in time with a KPM expansion. (c) Pablo Piskunow 2021.

Evolve an initial vector to arbitrary chosen points in time
(not necessarily small intervals), and compute expectation values of
observables.


"""

from functools import partial

import numpy as np
from scipy.special import jv

from .kpm_generator import GeneralVectorExpansion
from .kpm_generator import SpectralDensityIterator
from .kpm_generator import _BaseKPM


def coef_evo(m, e, ab, delta_time):
    """Return the coefficients of the time evolution operator.

    The 'delta_time' parameter must be the real time, and will be rescaled
    with respect to the bounds of the spectrum.

    The parameter 'e' is to conform with the signatures, but it is a
    dummy parameter, as the time evolution of a vector only depends on
    the Hamiltonian and the vector.
    """
    a, b = ab
    m_array = np.arange(m)
    num_e = len(np.atleast_1d(e).flatten())
    coef = jv(m_array, a * delta_time)
    coef = 2 * coef * ((-1j) ** m_array)
    coef[0] /= 2
    coef *= np.exp(-1j * b * delta_time)
    return np.array([coef for _ in range(num_e)]).T


def _max_num_moments(dt, accuracy):
    if dt == 0:
        return 1

    def func(dt, a, b, c, d, accu):
        return int(a * np.log(1 / accu) * (dt) ** b + c * dt + d * np.log(1 / accu) + 1)

    if accuracy < 1e-16:
        accu = 1e-32  # actual accuracy goal
        p = [0.33944873, 0.3304045, 1.0063549, 0.07014769, accu]
    elif accuracy < 1e-8:
        accu = 1e-16  # actual accuracy goal
        p = [0.27909421, 0.32633684, 1.02264908, 0.08207106, accu]
    elif accuracy < 1e-4:
        accu = 1e-8  # actual accuracy goal
        p = [0.23109766, 0.31671058, 1.0528007, 0.08506937, accu]
    else:
        raise ValueError(
            "This 'accuracy' is not very accurate, " "must be 'accuracy < 1e-4'."
        )

    return func(dt, *p)


def evolve_vectors(syst, vectors, dt=0, accuracy=1e-16, **kwargs):
    """Evolve vectors in time, according to a Hamiltonian.

    Parameters
    ----------
    syst : `~kwant.system` or Hamiltonian matrix
        Hamiltonian of the system.
    vectors : iterable of arrays
        Vectors to be passed to `~SpectralDensityIterator`.
    dt : float
        Time in units of E/hbar, where E is the units
        of energy of the Hamiltonian.
    accuracy : float, default to '1e-16'
        Accuracy goal for the Chebyshev expansion.
    kwargs : dict
        Extra keyword arguments to pass to `~SpectralDensityIterator`.
    """

    def identity(x):
        return x

    kwargs["kernel"] = identity
    kwargs["vector_factory"] = vectors
    num_vectors = len(vectors)
    kwargs["num_vectors"] = num_vectors

    bounds = kwargs.get("bounds", None)
    a = np.ptp(bounds) / 2
    # bound for the minimum number of moments
    rescaled_time = a * dt
    num_moments = _max_num_moments(rescaled_time, accuracy)
    kwargs["num_moments"] = num_moments

    gvec = GeneralVectorExpansion(
        syst, coef_function=partial(coef_evo, delta_time=dt), **kwargs
    )

    return gvec(0).squeeze(0)


def evolve_vectors_time_steps(
    syst, time_steps, vectors, save_vectors=True, accuracy=1e-16, **kwargs
):
    """Evolve vectors in time using time steps.

    Parameters
    ----------
    syst : `~kwant.system` or Hamiltonian matrix
        Hamiltonian of the system, where the vectors will evolve.
    time_steps : sequence of floats
        Discrete time steps that will be used to evolve the vectors.
    vectors : ndarray
        Initial sequence of vectors that belong to the system.
    save_vectors : bool, default to 'True'
        Whether to save the vectors at intermediate steps, and return them.
    accuracy : float, default to '1e-16'
        Precission of each time-evolution step.
    kwargs : dict
        Key-word arguments to be passed to 'evolve_vectors'
    """
    params = kwargs.get("params", dict())
    bounds = kwargs.get("bounds", None)
    if bounds is None:
        base = _BaseKPM(syst, params=params, num_moments=2, num_vectors=1)
        bounds = base.bounds

    dts = np.ediff1d(time_steps)
    v_t = vectors

    final_v = [v_t]
    for dt in dts:
        v_t = evolve_vectors(
            syst, vectors=v_t, dt=dt, accuracy=accuracy, bounds=bounds, params=params
        )
        if save_vectors:
            final_v.append(v_t)
    if not save_vectors:
        final_v = v_t
    return np.array(final_v)


def evolve_observable(
    syst,
    time_steps,
    operator=None,
    vectors=None,
    accuracy=1e-16,
    save_intermediate=True,
    **kwargs
):
    """Evolve an observable in time using time steps.

    Parameters
    ----------
    syst : `~kwant.system` or Hamiltonian matrix
        Hamiltonian of the system, where the vectors will evolve.
    time_steps : sequence of floats
        Discrete time steps that will be used to evolve the vectors.
    vectors : ndarray
        Initial sequence of vectors that belong to the system.
    operator : callable, ndarray, or sparse matrix, optional
        Any operator that can be accepted by `SpectralDensityIterator`.
    accuracy : float, default to '1e-16'
        Precission of each time-evolution step.
    save_intermediate: bool, default to True
        Save intermediate vectors.
    kwargs : dict
        Key-word arguments to be passed to 'evolve_vectors'
    """
    kwargs["operator"] = operator
    kwargs.pop("vector_factory", None)

    # parameters to pass to the time evolution KPM expansion
    tkwargs = dict(bounds=kwargs.get("bounds"), params=kwargs.get("params"))

    dts = np.ediff1d(time_steps)
    v_t = vectors
    densities_t = []

    if save_intermediate:
        spectrum = SpectralDensityIterator(syst, vector_factory=v_t, **kwargs)
        densities_t.append(spectrum.densities)

    for dt in dts:
        v_t = evolve_vectors(syst, v_t, dt=dt, accuracy=accuracy, **tkwargs)
        if save_intermediate:
            spectrum = SpectralDensityIterator(syst, vector_factory=v_t, **kwargs)
            densities_t.append(spectrum.densities)

    if not save_intermediate:
        spectrum = SpectralDensityIterator(syst, vector_factory=v_t, **kwargs)
        densities_t.append(spectrum.densities)

    return densities_t



"""Quick reference of topological systems.

Defined both with open boundary and periodic boundary conditions.
"""
from itertools import product

import kwant
import numpy as np
from kwant import HoppingKind


# neighbors for honeycomb lattice
honeycomb = kwant.lattice.honeycomb(1, norbs=1, name=["a", "b"])
honeycomb_a, honeycomb_b = honeycomb.sublattices

first_neighbors = [
    HoppingKind((0, 1), honeycomb_a, honeycomb_b),
    HoppingKind((0, 0), honeycomb_a, honeycomb_b),
    HoppingKind((-1, 1), honeycomb_a, honeycomb_b),
]

second_neighbors_a = [
    HoppingKind((0, 1), honeycomb_a, honeycomb_a),
    HoppingKind((1, -1), honeycomb_a, honeycomb_a),
    HoppingKind((-1, 0), honeycomb_a, honeycomb_a),
]

second_neighbors_b = [
    HoppingKind((1, 0), honeycomb_b, honeycomb_b),
    HoppingKind((0, -1), honeycomb_b, honeycomb_b),
    HoppingKind((-1, 1), honeycomb_b, honeycomb_b),
]


def haldane_pbc(trans_symm=((1, 0), (0, 1)), e_0=0, t=1, t2=0.5, return_builder=False):
    """Build a Haldane system with periodic boundary conditions."""
    periods = np.dot(trans_symm, honeycomb.prim_vecs)
    syst = kwant.builder.Builder(kwant.lattice.TranslationalSymmetry(*periods))

    syst[honeycomb.shape(lambda s: True, (0, 0))] = e_0
    syst[first_neighbors] = t
    syst[second_neighbors_a] = 1j * t2
    syst[second_neighbors_b] = 1j * t2

    if return_builder:
        return honeycomb, syst
    return honeycomb, kwant.wraparound.wraparound(syst).finalized()


def haldane_obc(trans_symm=((1, 0), (0, 1)), e_0=0, t=1, t2=0.5):
    """Build a Haldane system with open boundary conditions."""
    trans_symm = np.array(trans_symm, dtype=int)
    indices_left = trans_symm[1] - trans_symm[0]
    sign = np.sign(indices_left)

    syst = kwant.builder.Builder()

    for i, j in product(
        range(0, indices_left[0], sign[0]), range(0, indices_left[1], sign[1])
    ):
        tag = trans_symm[0] + np.array([i, j], dtype=int)
        site_a = honeycomb_a(*tag)
        site_b = honeycomb_b(*tag)
        syst[site_a] = e_0
        syst[site_b] = e_0

    syst[first_neighbors] = t
    syst[second_neighbors_a] = 1j * t2
    syst[second_neighbors_b] = 1j * t2

    return honeycomb, syst.finalized()


params_qhe = dict(mu=-4, t=-1, B=np.pi / 7)


def qhe_obc(length=50, width=10, init_pos=(0, 0)):
    """Build a quantum Hall system with open boundary conditions."""

    def bar(pos):
        x = pos[0] - init_pos[0]
        y = pos[1] - init_pos[1]
        return (x >= 0 and x < length) and (y >= 0 and y < width)

    # Onsite and hoppings
    def onsite(site, t, mu):
        return 4 * t - mu

    def hopping_ax(site1, site2, t, B):  # noqa: N803
        xt, yt = site1.pos
        xs, ys = site2.pos
        return -t * np.exp(-0.5j * B * (xt + xs) * (yt - ys))

    # Building system
    lat = kwant.lattice.square(norbs=1)
    syst = kwant.builder.Builder()

    syst[lat.shape(bar, init_pos)] = onsite
    syst[lat.neighbors()] = hopping_ax

    return syst.finalized()


def qhe_pbc(trans_symm=((1, 0), (0, 1)), init_pos=(0, 0), return_builder=False):
    """Build a quantum Hall system with periodic boundary conditions."""
    # Building system
    lat = kwant.lattice.square(norbs=1)
    periods = np.dot(trans_symm, lat.prim_vecs)
    syst = kwant.builder.Builder(kwant.lattice.TranslationalSymmetry(*periods))

    # Onsite and hoppings
    def onsite(site, t, mu):
        return 4 * t - mu

    def hopping_ax(site1, site2, t, B):  # noqa: N803
        xt, yt = site1.pos
        xs, ys = site2.pos
        return -t * np.exp(-0.5j * B * (xt + xs) * (yt - ys))

    syst[lat.shape(lambda pos: True, init_pos)] = onsite
    syst[lat.neighbors()] = hopping_ax

    if return_builder:
        return syst
    return kwant.wraparound.wraparound(syst).finalized()



"""Generate tile vectors to compute the effective Bloch spectrum."""

from collections import defaultdict

import kwant
import numpy as np
from kwant._common import ensure_rng


def momentum_to_ts(k, ts):
    """Transform 'k' to the reciprocal lattice basis."""
    # columns of B are lattice vectors
    _b = np.array(ts.periods).T
    # columns of A are reciprocal lattice vectors
    _a = np.linalg.pinv(_b).T
    k, residuals = np.linalg.lstsq(_a, k, rcond=None)[:2]
    if np.any(abs(residuals) > 1e-7):
        raise RuntimeError(
            "Requested momentum doesn't correspond to "
            "any lattice momentum or TranslationalSymmetry."
        )
    return k


class _TileVectors:
    def __init__(self, sites, tile, ts, *, check=True):
        """Construct an iterator over the sites in 'tile'.

        The iterator returns one vector for every orbital for every site
        in the fundamental domain (FD) of 'tile'.
        Each vector contains the orbital and its copies on the other
        tiles.

        Parameters
        ----------
        sites : iterable of `kwant.builder.Site` instances
            The sites of the system. Should belong to a lattice that
            is commensurate with both 'tile' and 'ts'.
        tile :  instance of `~kwant.TranslationalSymmetry`
            Translational symmetry of the tiles within the supercell.
            The tiles should cover the supercell 'ts'.
        ts : instance of `~kwant.TranslationalSymmetry` or
            Translational symmetry of the supercell.
        check : bool, default to 'True'
            Check that 'tile' covers the 'ts'. This is done by checking
            that the translational symmetry 'tile' has the subgroup 'ts'.
            If 'False', do not check anything.

        """
        if check:
            # This does not yet checks
            # for lattice commensuration with tile nor TS.
            # The check is done when calling `tile.to_fd()`.
            if not tile.has_subgroup(ts):
                raise ValueError(f"The tile {tile} has no subgroup {ts}.")

        self.tile = tile
        self.ts = ts

        # assume all norbs are the same
        self.norbs = sites[0].family.norbs
        self.tot_norbs = self.norbs * len(sites)

        self.site_replicas = defaultdict(list)
        for idx, site in enumerate(sites):
            base = tile.to_fd(site)
            self.site_replicas[base].append(idx)

        self.keys = list(self.site_replicas.keys())

        # these counters can be reset to default values
        self.idx = 0
        self.orb_idx = 0

    def _reset_count(self):
        self.idx = 0
        self.orb_idx = 0

    def __len__(self):
        return self.norbs * len(self.keys)

    def __iter__(self):
        return self

    @property
    def _norm(self):
        return 1 / len(self)

    def __next__(self):
        try:
            keys = self.keys[self.idx]
        except IndexError as err:
            raise StopIteration(
                "Too many vectors requested " "from this generator."
            ) from err

        site_indices = np.array(self.site_replicas[keys], dtype=int)
        orbs = site_indices * self.norbs + self.orb_idx

        self.orb_idx += 1
        # bump indices if needed
        if self.orb_idx >= self.norbs:
            self.idx += 1
            self.orb_idx = 0

        return site_indices, orbs


class TileKVectors(_TileVectors):
    """Construct k-vectors over the sites in 'tile'.

    The iterator returns one vector for every orbital for every site
    in the fundamental domain (FD) of 'tile'.
    Each vector contains the orbital and its copies on the other
    tiles, multiplied with the phase given by the position and the
    'k' vector.

    Parameters
    ----------
    k : pair or triplet floats
        The vector in k-space. Should have the same dimensions as the
        positions of the 'syst'.
    tile :  instance of `~kwant.TranslationalSymmetry`
        Translational symmetry of the tiles within the supercell.
        The tiles should cover the supercell defined by the wrapped
        symmetry of 'syst'.
    syst : instance of `~kwant.system.FiniteSystem`
        System with wrapped symmetry.

    Note
    ----
    'kwargs' are passed to `_TileVectors`.

    """

    def __init__(self, k, tile, syst, **kwargs):
        """Initialize tiles."""
        self.k = k
        self.positions = np.array([s.pos for s in syst.sites])

        if isinstance(syst.symmetry, kwant.builder.NoSymmetry):
            ts = syst._wrapped_symmetry
        else:
            ts = syst.symmetry

        super().__init__(syst.sites, tile, ts, **kwargs)

    def _get_orbs_and_phase(self):
        site_indices, orbs = super().__next__()
        pos = self.positions[site_indices]
        phase = np.exp(1j * pos.dot(self.k))
        return orbs, phase

    def __next__(
        self,
    ):
        """Get the next vector."""
        orbs, phase = self._get_orbs_and_phase()

        vector = np.zeros(self.tot_norbs, dtype=complex)
        vector[orbs] = phase
        return vector


def tile_random_kvectors(k, tile, syst, *, rng=0, **kwargs):
    """Construct an iterator of random phase vectors inside 'tile'.

    The other copies of 'tile' will have a phase relative to the base
    tile. That phase is given by 'k' and the positions of the sites in
    the 'syst'. Check 'TileKVectors' for details.

    Parameters
    ----------
    k : pair or triplet floats
        The vector in k-space. Should have the same dimensions as the
        positions of the 'syst'.
    tile :  instance of `~kwant.TranslationalSymmetry`
        Translational symmetry of the tiles within the supercell.
        The tiles should cover the supercell defined by the wrapped
        symmetry of 'syst'.
    syst : instance of `~kwant.system.FiniteSystem`
        System with wrapped symmetry.
    rng : int, float, or string, default to 'int(0)'
        Initial seed for the random vectors to ensure reproducibility.

    Note
    ----
    'kwargs' are passed to `_TileVectors`.
    """
    rng = ensure_rng(rng)
    tile_kvecs = TileKVectors(k, tile, syst, **kwargs)

    while True:
        tile_kvecs._reset_count()
        vector = np.zeros(tile_kvecs.tot_norbs, dtype=complex)
        for _ in range(len(tile_kvecs)):
            orbs, phase = tile_kvecs._get_orbs_and_phase()
            vector[orbs] = phase * np.exp(2j * np.pi * rng.random_sample())
        yield vector



"""Plotting functions and others."""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

from .common import trace_unit_cell


def plot_kwant(
    sites,
    site_color=None,
    linewidths=0.2,
    site_size=10,
    edgecolor="k",
    cmap="Blues",
    depth=False,
    elev=None,
    azim=None,
    ax=None,
    show=False,
    unit_cell=False,
    pos_transform=None,
    colorbar=True,
    **kwargs
):
    """Plot a list of kwant system sites with site colors and colorbar."""
    if site_color is None:
        site_color = np.ones(len(sites))
    if unit_cell:
        family_name = sites[0].family.name
        sites = [s for s in sites if s.family.name == family_name]
        site_color = trace_unit_cell(sites, site_color)
        # traces norbs and sublattices
    else:
        sites = sites
        site_color = site_color
    if len(sites) != site_color.shape[0]:
        raise ValueError(
            "The number of sites (orbitals) to plot must match " "the 'array' length"
        )

    dim = sites[0].family.dim

    fig = plt.gcf()

    if pos_transform is None:

        def pos_transform(x):
            return x

    positions = np.array([pos_transform(s.pos) for s in sites]).T

    if site_color is None:
        site_color = np.ones(positions.shape[1])

    return_artists = True
    if ax is None:
        return_artists = False
        if dim == 3:
            ax = fig.add_subplot(111, projection="3d")
        else:
            ax = fig.add_subplot(111)

    # TODO: check *positions warning
    if dim == 3:
        art = ax.scatter3D(
            *positions,
            c=site_color,
            cmap=cmap,
            linewidths=linewidths,
            s=site_size,
            edgecolor=edgecolor,
            depthshade=depth,
            **kwargs
        )

        ax.view_init(elev=elev, azim=azim)
    else:
        art = ax.scatter(
            *positions,
            c=site_color,
            cmap=cmap,
            linewidths=linewidths,
            s=site_size,
            edgecolor=edgecolor,
            **kwargs
        )

    if colorbar:
        fig.colorbar(art, ax=ax)

    if show:
        fig.show()

    if return_artists:
        return art
    return fig


def plot_neighbors(
    syst,
    lat=None,
    neighs=None,
    show=False,
    more_sites=False,
    init_tags=((0, 0)),
    ax=None,
    scatter_kwargs=None,
    arrow_kwargs=None,
):
    """Plot arrows between the neighbors."""
    if scatter_kwargs is None:
        scatter_kwargs = {}
    if arrow_kwargs is None:
        arrow_kwargs = {}
    plot_kwargs = dict(
        color="k", arrowstyle="simple", mutation_scale=15, lw=0, alpha=0.5
    )
    plot_kwargs.update(arrow_kwargs)
    dim = lat.sublattices[0].dim
    if neighs is None:
        if lat is None:
            try:
                lat = syst.sites[0].family
            except Exception:
                lat = list(syst.sites())[0].family
        neighs = []
        for sublat in lat.sublattices:
            neighs.extend(sublat.neighbors())

    if ax is None:
        ax = plt.gca()

    for init_tag in init_tags:
        for nei in neighs:
            site_a = nei.family_a(*init_tag)
            second_tag = np.array(init_tag) - np.array(nei.delta)
            site_b = nei.family_b(*(second_tag))
            delta_pos = site_b.pos - site_a.pos

            if dim == 2:
                ax.arrow(
                    x=site_a.pos[0],
                    y=site_a.pos[1],
                    dx=delta_pos[0],
                    dy=delta_pos[1],
                    head_width=0.05,
                    head_length=0.1,
                    fc="k",
                    ec="k",
                )
            elif dim == 3:
                rs = np.array([site_a.pos, site_a.pos + delta_pos]).T
                arrow = Arrow3D(*rs, **plot_kwargs)
                ax.add_artist(arrow)

            ax.scatter(*site_a.pos, **scatter_kwargs)
            ax.scatter(*site_b.pos, **scatter_kwargs)

            if more_sites:
                site_a = nei.family_a(*second_tag)
                second_tag = second_tag - np.array(nei.delta)
                site_b = nei.family_b(*(second_tag))
                delta_pos = site_b.pos - site_a.pos

                if dim == 2:
                    ax.arrow(
                        x=site_a.pos[0],
                        y=site_a.pos[1],
                        dx=delta_pos[0],
                        dy=delta_pos[1],
                        head_width=0.05,
                        head_length=0.1,
                        fc="k",
                        ec="k",
                    )
                elif dim == 3:
                    rs = np.array([site_a.pos, site_a.pos + delta_pos]).T
                    arrow = Arrow3D(*rs, **plot_kwargs)
                    ax.add_artist(arrow)
                ax.scatter(*site_a.pos, **scatter_kwargs)
                ax.scatter(*site_b.pos, **scatter_kwargs)
    if show:
        plt.show()
    return ax


def plot_vectors(
    vectors,
    ax=None,
    hl=0.1,  # head_length
    head_width=0.05,
    color="k",  # can be overridden,
    show=False,
    fig_size=None,
    resize_ax=False,
    margin=0.2,
    **kwargs
):
    """Plot vectors.

    Args:
        vectors (_type_): _description_
        ax (_type_, optional): _description_. Defaults to None.
        hl (float, optional): _description_. Defaults to 0.1.
        color (str, optional): _description_. Defaults to "k".
        show (bool, optional): _description_. Defaults to False.
        fig_size (_type_, optional): _description_. Defaults to None.
        resize_ax (bool, optional): _description_. Defaults to False.
        margin (float, optional): _description_. Defaults to 0.2.

    Returns:
        _type_: _description_
    """
    fig = plt.gcf()
    if fig_size is not None:
        fig.set_size_inches(fig_size)
    if ax is None:
        ax = plt.gca()

    directions = []
    for vector in vectors:
        _ec = vector.get("color", color)
        _fc = vector.get("color", color)
        _hl = vector.get("hl", hl)
        _hw = vector.get("head_width", head_width)
        origin = vector.get("origin", (0, 0))
        direction = vector.get("direction", (1, 1))
        directions.append(direction)

        ax.arrow(
            x=origin[0],
            y=origin[1],
            dx=direction[0],
            dy=direction[1],
            head_width=_hw,
            head_length=_hl,
            fc=_fc,
            ec=_ec,
            **kwargs
        )

    directions = np.array(directions)

    if resize_ax:
        ptp = np.ptp(directions, axis=0) * margin  # % margin
        lower_bound = np.min(directions, axis=0)
        upper_bound = np.max(directions, axis=0)

        ax.set_xlim(lower_bound[0] - ptp[0], upper_bound[0] + ptp[0])
        ax.set_ylim(lower_bound[1] - ptp[1], upper_bound[1] + ptp[1])
    if show:
        plt.show()
    return fig


class Arrow3D(FancyArrowPatch):
    """Patch fancy arrow.

    Args:
        FancyArrowPatch (_type_): _description_
    """

    def __init__(self, xs, ys, zs, *args, **kwargs):
        """Initialize.

        Args:
            xs (_type_): _description_
            ys (_type_): _description_
            zs (_type_): _description_
        """
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        """Draw.

        Args:
            renderer (_type_): _description_
        """
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


def plot_arrows3d(start=None, arrow=None, ax=None, show=False, **kwargs):
    """Plot 3d arrows.

    Args:
        start (_type_, optional): _description_. Defaults to None.
        arrow (_type_, optional): _description_. Defaults to None.
        ax (_type_, optional): _description_. Defaults to None.
        show (bool, optional): _description_. Defaults to False.

    Returns:
        _type_: _description_
    """
    fig = plt.gcf()
    if ax is None:
        ax = fig.add_subplot(111, projection="3d")

    plot_kwargs = dict(mutation_scale=5, lw=3, arrowstyle="-|>", color="b", alpha=0.5)
    plot_kwargs.update(kwargs)

    for i in range(len(start)):
        a = Arrow3D(
            [start[i, 0], start[i, 0] + arrow[i, 0]],
            [start[i, 1], start[i, 1] + arrow[i, 1]],
            [start[i, 2], start[i, 2] + arrow[i, 2]],
            **plot_kwargs
        )
        ax.add_artist(a)

    if show:
        plt.show()
    return fig



"""KPM Tools."""
# flake8: noqa
from .bloch import _hopping_distance
from .bloch import separate_bloch_components
from .bloch import wrap_distance
from .bloch import wrap_velocity
from .bloch import wraparound_by_parts
from .common import expectation_value
from .common import identity_operator
from .common import local_expectation_value
from .common import position_operator
from .common import standard_gaussian
from .common import trace_orbs
from .common import trace_unit_cell
from .common import transparent_cmap
from .common import uniform
from .evolution import _max_num_moments
from .evolution import coef_evo
from .evolution import evolve_observable
from .evolution import evolve_vectors
from .evolution import evolve_vectors_time_steps
from .hamiltonians import haldane_obc
from .hamiltonians import haldane_pbc
from .hamiltonians import qhe_obc
from .hamiltonians import qhe_pbc
from .kpm_funcs import chebysev_polynomial_series
from .kpm_funcs import chern_number
from .kpm_funcs import chern_number_pbc
from .kpm_funcs import coef_delta_function
from .kpm_funcs import coef_greens_function
from .kpm_funcs import coef_greens_function_a
from .kpm_funcs import coef_greens_function_prime
from .kpm_funcs import coef_greens_function_prime_a
from .kpm_funcs import coef_greens_function_prime_r
from .kpm_funcs import coef_greens_function_r
from .kpm_funcs import coef_projector_function
from .kpm_funcs import coef_projector_function_unoccupied
from .kpm_funcs import conductivity
from .kpm_funcs import delta_function
from .kpm_funcs import greens_function
from .kpm_funcs import longitudinal_conductivity
from .kpm_funcs import projector_operator
from .kpm_generator import GeneralVectorExpansion
from .kpm_generator import IteratorKPM
from .kpm_generator import SpectralDensityIterator
from .kpm_generator import _BaseKPM
from .kpm_generator import _dagger
from .kpm_generator import _normalize_num_moments
from .kpm_generator import _normalize_operator
from .kpm_generator import _rescale
from .kpm_generator import concatenator
from .kpm_generator import general_function
from .kpm_generator import multiprocess_update
from .kpm_generator import process_update
from .kpm_generator import thread_update
from .plotting import Arrow3D
from .plotting import plot_arrows3d
from .plotting import plot_kwant
from .plotting import plot_neighbors
from .plotting import plot_vectors
from .tiles import TileKVectors
from .tiles import _TileVectors
from .tiles import momentum_to_ts
from .tiles import tile_random_kvectors



"""Pre-built common functions using the KPM generators."""
import numpy as np

from .common import identity_operator
from .common import position_operator
from .kpm_generator import concatenator
from .kpm_generator import general_function


def greens_function(syst, params, **kwargs):
    """Build a Green's function operator using KPM.

    Returns a function that takes an energy or a list of energies, and returns
    the Green's function with that energy acting on `vectors`.

    Parameters
    ----------
    syst : kwant.System or ndarray
        Finalized kwant system or dense or sparse ndarray of the
        Hamiltonian with shape `(N, N)`.
    params : dict, optional
        Parameters for the kwant system.
    kpm_params : dict, optional
        Dictionary containing the parameters to pass to the `~kwant.kpm`
        module. 'num_vectors' will be overwritten to match the number
        of vectors, and 'operator' key will be deleted.
    precalculate_moments: bool, default False
        Whether to precalculate and store all the KPM moments of `vectors`.
        This is useful if the Green's function is evaluated at a large
        number of energies, but uses a large amount of memory.
        If False, the KPM expansion is performed every time the Green's
        function is called, which minimizes memory use.

    Returns
    -------
    green_expansion : callable
        Takes an energy or array of energies and returns the Greens function
        acting on the vectors, for those energies.

    """
    return general_function(syst, params, coef_function=coef_greens_function, **kwargs)


def delta_function(syst, params, **kwargs):
    """Build a projector over the occupied energies.

    Returns a function that takes a Fermi energy, and returns the
    projection of the `vectors` over the occupied energies of the
    Hamiltonian.

    Parameters
    ----------
    syst : kwant.System or ndarray
        Finalized kwant system or dense or sparse ndarray of the
        Hamiltonian with shape `(N, N)`.
    params : dict, optional
        Parameters for the kwant system.
    kpm_params : dict, optional
        Dictionary containing the parameters to pass to the `~kwant.kpm`
        module. 'num_vectors' will be overwritten to match the number
        of vectors, and 'operator' key will be deleted.
    precalculate_moments: bool, default False
        Whether to precalculate and store all the KPM moments of `vectors`.
        This is useful if the Green's function is evaluated at a large
        number of energies, but uses a large amount of memory.
        If False, the KPM expansion is performed every time the Green's
        function is called, which minimizes memory use.


    Returns
    -------
    projector : callable
        Takes an energy or array of energies and returns the projection
        onto the occupied states of the vectors, for those energies.

    """
    return general_function(syst, params, coef_function=coef_delta_function, **kwargs)


def projector_operator(syst, params, **kwargs):
    """Build a projector over the occupied energies.

    Returns a function that takes a Fermi energy, and returns the
    projection of the `vectors` over the occupied energies of the
    Hamiltonian.

    Parameters
    ----------
    syst : kwant.System or ndarray
        Finalized kwant system or dense or sparse ndarray of the
        Hamiltonian with shape `(N, N)`.
    params : dict, optional
        Parameters for the kwant system.
    kpm_params : dict, optional
        Dictionary containing the parameters to pass to the `~kwant.kpm`
        module. 'num_vectors' will be overwritten to match the number
        of vectors, and 'operator' key will be deleted.
    precalculate_moments: bool, default False
        Whether to precalculate and store all the KPM moments of `vectors`.
        This is useful if the Green's function is evaluated at a large
        number of energies, but uses a large amount of memory.
        If False, the KPM expansion is performed every time the Green's
        function is called, which minimizes memory use.


    Returns
    -------
    projector : callable
        Takes an energy or array of energies and returns the projection
        onto the occupied states of the vectors, for those energies.

    """
    return general_function(
        syst, params, coef_function=coef_projector_function, **kwargs
    )


def coef_greens_function(m, e, sign=-1, ab=(1, 0)):
    """Generate Green's functions coefficients."""
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    prefactor = -2j * sign / (np.sqrt((1 - e_rescaled) * (1 + e_rescaled)) * a)
    phi_e = np.arccos(e_rescaled)

    m_array = np.arange(m)

    coef = np.exp(1j * sign * np.outer(m_array, phi_e))

    coef[0] = coef[0] / 2

    return prefactor * coef


def coef_greens_function_r(m, e, ab=(1, 0)):
    """Generate the retarded Green's functions coefficients."""
    return coef_greens_function(m, e, ab=ab, sign=-1)


def coef_greens_function_a(m, e, ab=(1, 0)):
    """Generate the advanced Green's functions coefficients."""
    return coef_greens_function(m, e, ab=ab, sign=1)


def coef_greens_function_prime(m, e, sign=-1, ab=(1, 0)):
    """Generate the prime Green's functions coefficients."""
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    one_e2 = (1 - e_rescaled) * (1 + e_rescaled)
    sqrt_e = np.sqrt(one_e2)
    prefactor = -2j * sign / (one_e2 * a**2)
    phi_e = np.arccos(e_rescaled)

    m_array = np.arange(m)

    coef = -1j * sign * m_array[:, np.newaxis]
    coef = coef + (e_rescaled / sqrt_e)[np.newaxis, :]
    coef = coef * np.exp(1j * sign * np.outer(m_array, phi_e))

    coef[0] = coef[0] / 2

    return prefactor * coef


def coef_greens_function_prime_r(m, e, ab=(1, 0)):
    """Generate the prime retarded Green's functions coefficients."""
    return coef_greens_function_prime(m, e, ab=ab, sign=-1)


def coef_greens_function_prime_a(m, e, ab=(1, 0)):
    """Generate the prime advanced Green's functions coefficients."""
    return coef_greens_function_prime(m, e, ab=ab, sign=1)


def coef_projector_function(m, e, ab=(1, 0)):
    """Projector over occupied states below the Fermi energy."""
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    m_array = np.arange(m)
    m_array[0] = 1
    phi_e = np.arccos(e_rescaled)
    coef = -2 / np.pi * np.sin(np.outer(phi_e, m_array)) / m_array
    coef[:, 0] = 1 - phi_e / np.pi
    return coef.T


def coef_projector_function_unoccupied(m, e, ab=(1, 0)):
    """Projector over unoccupied states below the Fermi energy."""
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    m_array = np.arange(m)
    m_array[0] = 1
    phi_e = np.arccos(e_rescaled)
    coef = 2 / np.pi * np.sin(np.outer(phi_e, m_array)) / m_array
    coef[:, 0] = phi_e / np.pi
    return coef.T


def coef_delta_function(m, e, ab=(1, 0)):
    """Generate the delta function coefficients."""
    coef = 2 * chebysev_polynomial_series(m, e, ab=ab)
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    g_e = (np.pi * np.sqrt(1 - e_rescaled) * np.sqrt(1 + e_rescaled)) * a

    return coef / g_e


def chebysev_polynomial_series(m, e, ab=(1, 0)):
    """Generate the Chebyshev polynomial series."""
    # the first coefficient is halved
    a, b = ab
    e_rescaled = (np.atleast_1d(e) - b) / a
    e2 = 2 * e_rescaled
    coef = np.ndarray((m, e_rescaled.shape[0]))

    coef[0] = np.ones_like(e_rescaled)
    if m > 1:
        coef[1] = e_rescaled
        for i in range(2, m):
            coef[i] = coef[i - 1] * e2 - coef[i - 2]

    coef[0] = coef[0] / 2
    return coef


def chern_number(syst, x=None, y=None, **kwargs):
    """Compute the Chern number in real space.

    Parameters
    ----------
    syst : kwant system or Hamiltonian matrix
        The system to compute the Chern number.
    x : callable, optional
        First (second) position operator. If provided, it
        must be the '.dot' method of an ndarray. If not
        provided, the 'x' and 'y' canonical coordinates
        extraxted from 'syst' will be used. In this case,
        'syst' must be a kwant system.
    y : callable, optional
        Same as `x`.
    kwargs : dict
        The remaining keyword arguments are passed to
        `~kpm_generator.concatenator`.

    """
    # make x and y callables
    if x is None and y is None:
        x_op, y_op = position_operator(syst)
        x = x_op.dot
        y = y_op.dot
    elif x is None:
        x_op, y_op = position_operator(syst)
        x = x_op.dot
    elif y is None:
        x_op, y_op = position_operator(syst)
        y = y_op.dot

    chern = concatenator(
        syst,
        operator_array=[None, x, y, None],
        coef_array=[coef_projector_function] * 3,
        first_transpose=False,
        ouroboros=True,
        **kwargs
    )

    return lambda e: 4 * np.pi * chern(e).imag


def chern_number_pbc(syst, **kwargs):
    """Generate the Chern number for periodic boundary conditions."""
    x_op, y_op = position_operator(syst)

    chern_1 = concatenator(
        syst,
        operator_array=[None, x_op.dot, y_op.dot, None],
        coef_array=[
            coef_projector_function,
            coef_projector_function_unoccupied,
            coef_projector_function,
        ],
        first_transpose=False,
        ouroboros=True,
        **kwargs
    )

    chern_2 = concatenator(
        syst,
        operator_array=[None, y_op.dot, x_op.dot, None],
        coef_array=[
            coef_projector_function,
            coef_projector_function_unoccupied,
            coef_projector_function,
        ],
        first_transpose=False,
        ouroboros=True,
        **kwargs
    )

    return lambda e: 2j * np.pi * (chern_1(e) - chern_2(e))


def conductivity(syst, op_a="x", op_b="y", **kwargs):
    """Compute the Kubo-Bastin conductivity kernel.

    The result must be normalized with the area per site,
    and integrated up to the Fermi energy to obtain the Kubo-Bastin
    conductivity.
    """
    from kwant.kpm import _velocity

    params = kwargs.get("params", None)
    positions = kwargs.get("positions", None)

    v_a = _velocity(syst, params, op_a, positions=positions)
    v_b = _velocity(syst, params, op_b, positions=positions)

    kb_1 = concatenator(
        syst,
        operator_array=[v_a.dot, v_b.dot, identity_operator],
        coef_array=[coef_delta_function, coef_greens_function_prime_a],
        first_transpose=True,
        ouroboros=False,
        **kwargs
    )

    kb_2 = concatenator(
        syst,
        operator_array=[v_a.dot, v_b.dot, identity_operator],
        coef_array=[coef_greens_function_prime_r, coef_delta_function],
        first_transpose=True,
        ouroboros=False,
        **kwargs
    )

    return lambda e: 2 * np.pi * np.real(1j * (kb_1(e) - kb_2(e)))


def longitudinal_conductivity(syst, direction="x", **kwargs):
    """Compute the Kubo-Greenwood conductivity.

    The result must be normalized with the area per site.
    The Kubo-Greenwood KPM expansion uses of the velocity operator.
    This operator can be either provided or computed inside this function,
    check the 'direction' parameter.

    Parameters
    ----------
    syst: ndarray or a Kwant system
        System for which the velocity operator is calculated.
    params : dict
        Parametres of the system
    direction: str, matrix or operator
        If 'direction' is a `str` in {'x', 'y', 'z'}, the velocity operator
        is calculated using the 'syst' and 'positions', else
        if 'direction' is an operator or a matrix, then that is the velocity
        operator.
    positions : ndarray of shape '(N, dim)', optional
        Positions of each orbital. This parameter is not used if
        'syst' is a Kwant system.
    """
    from kwant.kpm import _velocity

    params = kwargs.get("params", None)
    positions = kwargs.get("positions", None)

    v_a = _velocity(syst, params, direction, positions=positions)

    kg = concatenator(
        syst,
        operator_array=[v_a.dot, v_a.dot, identity_operator],
        coef_array=[coef_delta_function, coef_delta_function],
        first_transpose=True,
        ouroboros=False,
        **kwargs
    )

    return kg



"""Handle systems with periodic boundary conditions.

'wraparound_by_parts' is reworked from `~kwant.wraparound`, specially MR !363.
"""
import cmath
import collections
import inspect
import warnings

import numpy as np
import tinyarray as ta
from kwant._common import get_parameters
from kwant._common import memoize
from kwant.builder import Builder
from kwant.builder import herm_conj
from kwant.lattice import TranslationalSymmetry
from kwant.system import SiteArray
from kwant.wraparound import WrappedBuilder
from kwant.wraparound import _callable_herm_conj
from kwant.wraparound import _set_signature
from kwant.wraparound import wraparound


def wraparound_by_parts(
    builder, keep=None, *, coordinate_names="xyz", separate_sites=False
):
    """Replace translational symmetries by momentum parameters.

    A new Builder instance is returned.  By default, each symmetry is replaced
    by one scalar momentum parameter that is appended to the already existing
    arguments of the system.  Optionally, one symmetry may be kept by using the
    `keep` argument. The momentum parameters will have names like 'k_n' where
    the 'n' are specified by 'coordinate_names'.

    Parameters
    ----------
    builder : `~kwant.builder.Builder`
        Kwant builder with translational symmetries.
    keep : int, optional
        Which (if any) translational symmetry to keep.
    coordinate_names : sequence of strings, default: 'xyz'
        The names of the coordinates along the symmetry
        directions of 'builder'.
    separate_sites : bool, default to 'False'
        Wheather to construct a new system without the hoppings that cross
        the fundamental domain of the translational symmetry. If 'True'
        return the system contained inside the fundamental domain, a
        dictionary of wrapped sites and a dictionary of wrapped hoppings.

    Notes
    -----
    Wraparound is stop-gap functionality until Kwant 2.x which will include
    support for higher-dimension translational symmetry in the low-level system
    format. It will be deprecated in the 2.0 release of Kwant.

    """

    @memoize
    def bind_site(val):
        def f(*args):
            a, *args = args
            return val(a, *args[:mnp])

        if not callable(val):
            raise ValueError("Value function should be a Callable.")
        _set_signature(f, get_parameters(val) + momenta)
        return f

    @memoize
    def bind_hopping_as_site(elem, val):
        def f(*args):
            a, *args = args
            phase = cmath.exp(1j * ta.dot(elem, args[mnp:]))
            v = val(a, sym.act(elem, a), *args[:mnp]) if callable(val) else val
            pv = phase * v
            return pv + herm_conj(pv)

        params = ("_site0",)
        if callable(val):
            params += get_parameters(val)[2:]  # cut off both site parameters
        _set_signature(f, params + momenta)
        return f

    @memoize
    def bind_hopping(elem, val):
        def f(*args):
            a, b, *args = args
            phase = cmath.exp(1j * ta.dot(elem, args[mnp:]))
            v = val(a, sym.act(elem, b), *args[:mnp]) if callable(val) else val
            return phase * v

        params = ("_site0", "_site1")
        if callable(val):
            params += get_parameters(val)[2:]  # cut off site parameters
        _set_signature(f, params + momenta)
        return f

    @memoize
    def bind_sum(num_sites, *vals):
        """Construct joint signature for all 'vals'."""

        def f(*in_args):
            acc = 0
            for val, selection in val_selection_pairs:
                if selection:  # Otherwise: reuse previous out_args.
                    out_args = tuple(in_args[i] for i in selection)
                if callable(val):
                    acc = acc + val(*out_args)
                else:
                    acc = acc + val
            return acc

        params = collections.OrderedDict()

        # Add the leading one or two 'site' parameters.
        site_params = [f"_site{i}" for i in range(num_sites)]
        for name in site_params:
            params[name] = inspect.Parameter(name, inspect.Parameter.POSITIONAL_ONLY)

        # Add all the other parameters (except for the momenta).  Setup the
        # 'selections'.
        selections = []
        for val in vals:
            if not callable(val):
                selections.append(())
                continue
            val_params = get_parameters(val)[num_sites:]
            if val_params[mnp:] != momenta:
                raise ValueError("Shapes don't match.")
            val_params = val_params[:mnp]
            selections.append((*site_params, *val_params))
            for p in val_params:
                # Skip parameters that exist in previously added functions.
                if p in params:
                    continue
                params[p] = inspect.Parameter(p, inspect.Parameter.POSITIONAL_ONLY)

        # Sort values such that ones with the same arguments are bunched.
        # Prepare 'val_selection_pairs' that is used in the function 'f' above.
        params_keys = list(params.keys())
        val_selection_pairs = []
        prev_selection = None
        argsort = sorted(range(len(selections)), key=selections.__getitem__)
        momenta_sel = tuple(range(mnp, 0, 1))
        for i in argsort:
            selection = selections[i]
            if selection and selection != prev_selection:
                prev_selection = selection = (
                    tuple(params_keys.index(s) for s in selection) + momenta_sel
                )
            else:
                selection = ()
            val_selection_pairs.append((vals[i], selection))

        # Finally, add the momenta.
        for k in momenta:
            params[k] = inspect.Parameter(k, inspect.Parameter.POSITIONAL_ONLY)

        f.__signature__ = inspect.Signature(params.values())
        return f

    try:
        momenta = [
            f"k_{coordinate_names[i]}" for i in range(len(builder.symmetry.periods))
        ]
    except IndexError as err:
        raise ValueError(
            "All symmetry directions must have a name specified " "in coordinate_names"
        ) from err

    if keep is None:
        ret = WrappedBuilder()
        sym = builder.symmetry
    else:
        periods = list(builder.symmetry.periods)
        ret = WrappedBuilder(TranslationalSymmetry(periods.pop(keep)))
        sym = TranslationalSymmetry(*periods)
        momenta.pop(keep)
    momenta = tuple(momenta)
    mnp = -len(momenta)  # Used by the bound functions above.

    # Store the names of the momentum parameters and the symmetry of the
    # old Builder (this will be needed for band structure plotting)
    ret._momentum_names = momenta
    ret._wrapped_symmetry = builder.symmetry

    # Wrapped around system retains conservation law and chiral symmetry.
    # We use 'bind_site' to add the momenta arguments if required.
    cons = builder.conservation_law
    ret.conservation_law = bind_site(cons) if callable(cons) else cons
    chiral = builder.chiral
    ret.chiral = bind_site(chiral) if callable(chiral) else chiral

    if builder.particle_hole is not None or builder.time_reversal is not None:
        warnings.warn(
            "`particle_hole` and `time_reversal` symmetries are set "
            "on the input builder. However they are ignored for the "
            "wrapped system, since Kwant lacks a way to express the "
            "existence (or not) of a symmetry at k != 0.",
            RuntimeWarning,
            stacklevel=2,
        )
    ret.particle_hole = None
    ret.time_reversal = None

    ret.vectorize = builder.vectorize

    sites = {}
    hops = collections.defaultdict(list)

    # Store lists of values, so that multiple values can be assigned to the
    # same site or hopping.
    for site, val in builder.site_value_pairs():
        # Every 'site' is in the FD of the original symmetry.
        # Move the sites to the FD of the remaining symmetry, this guarantees that
        # every site in the new system is an image of an original FD site translated
        # purely by the remaining symmetry.
        sites[ret.symmetry.to_fd(site)] = [val]  # a list to append wrapped hoppings

    for hop, val in builder.hopping_value_pairs():
        a, b = hop
        # 'a' is in the FD of original symmetry.
        # Get translation from FD of original symmetry to 'b',
        # this is different from 'b_dom = sym.which(b)'.
        b_dom = builder.symmetry.which(b)
        # Throw away part that is in the remaining translation direction, so we get
        # an element of 'sym' which is being wrapped
        b_dom = ta.array([t for i, t in enumerate(b_dom) if i != keep])
        # Pull back using the remainder, which is purely in the wrapped directions.
        # This guarantees that 'b_wa' is an image of an original FD site translated
        # purely by the remaining symmetry.
        b_wa = sym.act(-b_dom, b)
        # Move the hopping to the FD of the remaining symmetry
        a, b_wa = ret.symmetry.to_fd(a, b_wa)

        if a == b_wa:
            # The hopping gets wrapped-around into an onsite Hamiltonian.
            # Since site `a` already exists in the system, we can simply append.
            sites[a].append(bind_hopping_as_site(b_dom, val))
        else:
            # The hopping remains a hopping.
            if any(b_dom):
                # The hopping got wrapped-around.
                val = bind_hopping(b_dom, val)

            # Make sure that there is only one entry for each hopping
            # pointing in one direction, modulo the remaining translations.
            b_wa_r, a_r = ret.symmetry.to_fd(b_wa, a)
            if (b_wa_r, a_r) in hops:
                if (a, b_wa) in hops:
                    raise ValueError("More than one entry per hopping.")

                if callable(val):
                    val = _callable_herm_conj(val)
                else:
                    val = herm_conj(val)

                hops[b_wa_r, a_r].append((val, b_dom))
            else:
                hops[a, b_wa].append((val, b_dom))

    # Copy stuff into result builder, converting lists of more than one element
    # into summing functions.

    # Optionally, store the wrapped sites and hoppings, and return them
    # together with the builder stripped of these sites.

    wrapped_sites = {}
    wrapped_hoppings = {}
    for site, vals in sites.items():
        if len(vals) == 1:
            # no need to bind onsites without extra wrapped hoppings
            ret[site] = vals[0]
        else:
            if separate_sites:
                ret[site] = vals[0]  # add the site without the wrapped part
                # conjugate is included by 'bind_sum'
                wrapped_sites[site] = bind_sum(1, *vals[1:])
            else:
                val = vals[0]
                vals[0] = bind_site(val) if callable(val) else val
                ret[site] = bind_sum(1, *vals)

    for hop, vals_doms in hops.items():
        if len(vals_doms) == 1:
            # no need to bind hoppings that are not already bound
            val, b_dom = vals_doms[0]

            # check if this hoppings is purely from wrapped translational symm.
            if any(b_dom) and separate_sites:
                wrapped_hoppings[hop] = val
            else:
                ret[hop] = val
        else:
            new_vals = [
                bind_hopping(b_dom, val)
                if callable(val) and not any(b_dom)  # skip hoppings already bound
                else val
                for val, b_dom in vals_doms
            ]
            if separate_sites:
                # add the hoppings that are not wrapped
                vals_in_fd = [val for val, b_dom in vals_doms if not any(b_dom)]
                if len(vals_in_fd) != 1:
                    raise ValueError("There should be only one hopping.")
                ret[hop] = vals_in_fd[0]

                vals_out_fd = [
                    val for val, b_dom in vals_doms if any(b_dom)
                ]  # vals are already bound

                wrapped_hoppings[hop] = bind_sum(2, *vals_out_fd)
            else:
                ret[hop] = bind_sum(2, *new_vals)

    if separate_sites:
        return ret, wrapped_sites, wrapped_hoppings
    return ret


def separate_bloch_components(builder):
    """Construct two systems: the supercell onsite and supercell hopping."""
    # get wrapped sites and hoppings separately
    syst_superonsite, w_sites, w_hoppings = wraparound_by_parts(
        builder, separate_sites=True
    )
    fsyst_bloch_superonsite = syst_superonsite.finalized()

    bloch_builder = Builder(symmetry=builder.symmetry)

    # the sites that are in the builder but not in the wrapped sites
    # should be added with zero value
    for s in builder.sites():
        norbs = s.family.norbs
        bloch_builder[s] = w_sites.get(s, np.zeros((norbs, norbs)))  # val or 0
    for hop, val in w_hoppings.items():
        bloch_builder[hop] = val
    fsyst_bloch_superhopping = wraparound(bloch_builder).finalized()

    return fsyst_bloch_superonsite, fsyst_bloch_superhopping


def _hopping_distance(site1, site2, direction):
    norbs = site1.family.norbs
    if norbs != site2.family.norbs:
        raise NotImplementedError(
            "Only hopppings between sites of equal number of orbitals is implemented."
        )

    if isinstance(site1, SiteArray):
        pos1 = site1.positions().transpose()
        pos2 = site2.positions().transpose()
        d = np.dot(direction, (pos1 - pos2))[:, np.newaxis, np.newaxis]
    else:
        pos1 = site1.pos
        pos2 = site2.pos
        d = np.dot(direction, (pos1 - pos2))

    # return an imaginary number so that the matrix is antisymmetric
    # but hermitian
    return 1j * d * np.identity(norbs)


def wrap_velocity(builder):
    """Construct velocity operators with translational symmetries.

    The system has an extra parameter 'direction' that should be a unit
    lenght vector in real space, of the dimensions of the site positions.

    The method 'hamiltonian_submatrix' produces the velocity operator matrix
    in the 'direction' specified.

    Note that this matrix depends on the value that takes the
    Hamiltonian of the original system, the 'builder'.
    """
    direction = ("direction",)
    dnp = -len(direction)

    @memoize
    def bind_velocity_hopping(val):
        def f(*args):
            a, b, *args = args  # first two args are sites
            d = _hopping_distance(a, b, *args[dnp:])
            # d is a diagonal of size norbs times a constant
            v = val(a, b, *args[:dnp]) if callable(val) else val
            v = np.atleast_1d(v)  # shape scalars
            return v @ d.swapaxes(-1, -2)  # transpose last two axes

        params = ("_site0", "_site1")
        if callable(val):
            params += get_parameters(val)[2:]  # cut off site parameters
        _set_signature(f, params + direction)
        return f

    velocity_builder = Builder(builder.symmetry, vectorize=builder.vectorize)

    for s in builder.sites():
        norbs = s.family.norbs
        velocity_builder[s] = np.zeros((norbs, norbs))

    for hop, val in builder.hopping_value_pairs():
        velocity_builder[hop] = bind_velocity_hopping(val)

    fsyst_velocity = wraparound(velocity_builder).finalized()

    return fsyst_velocity


def wrap_distance(builder):
    """Construct distance operators with translational symmetries.

    The system has a single parameter 'direction' that should be a unit
    lenght vector in real space, of the dimensions of the site positions.

    The method 'hamiltonian_submatrix' produces the distance operator matrix
    in the 'direction' specified.

    Note that this matrix is independent of the value that takes the
    Hamiltonian of the original system, and does not depends on the
    paramters of the original system, the 'builder'.
    """
    distance_builder = Builder(builder.symmetry, vectorize=builder.vectorize)

    for s in builder.sites():
        norbs = s.family.norbs
        distance_builder[s] = np.zeros((norbs, norbs))

    for hop, _ in builder.hopping_value_pairs():
        distance_builder[hop] = _hopping_distance

    fsyst_distance = wraparound(distance_builder).finalized()

    return fsyst_distance



"""Utils."""

from collections import OrderedDict

import kwant
import numpy as np
from matplotlib.colors import ListedColormap
from matplotlib.pylab import cm
from scipy.sparse import diags


def identity_operator(x):
    """Return argument."""
    return x


def position_operator(syst, pos_transform=None):
    """Return a list of position operators as 'csr' matrices."""
    operators = []
    norbs = syst.sites[0].family.norbs
    if pos_transform is None:
        pos = np.array([s.pos for s in syst.sites])
    else:
        pos = np.array([pos_transform(s.pos) for s in syst.sites])

    for i in range(pos.shape[1]):
        operators.append(diags(np.repeat(pos[:, i], norbs), format="csr"))
    return operators


def trace_orbs(array, norbs=4, axis=-1):
    """Return the trace per unit-cell.

    'array' has a length 'num_sites * norbs' over the 'axis',
    and the return array has a length 'num_sites' over the
    same axis.
    """
    array = np.asarray(array)
    if array.shape[axis] % norbs != 0:
        raise ValueError(f"The shape along axis {axis} must have shape {norbs}.")

    slices = [[slice(None) for i in range(array.ndim)] for i in range(norbs)]
    for n in range(norbs):
        slices[n][axis] = slice(n, None, norbs)
        slices[n] = tuple(slices[n])
    return sum([array[slices[n]] for n in range(norbs)])


def trace_unit_cell(sites, array):
    """Trace the array assuming it has the values per orbital per site.

    If
    value_per_orbital_per_site
    """
    norbs = sites[0].family.norbs
    value_per_site = trace_orbs(array, norbs, axis=0)
    norbs = sites[0].family.norbs

    trace = OrderedDict()
    for s, value in zip(sites, value_per_site, strict=True):
        _previous_value = trace.get(s.tag, 0)
        trace[s.tag] = _previous_value + value
    value_per_uc = np.array(list(trace.values()))

    return value_per_uc


def transparent_cmap(cmap=None, alpha="linear"):
    """Add transparency to a matplotlib ListedColormap.

    Parameters
    ----------
    cmap : instance of `matplotlib.pylab.cm`, optional
        The colormap to use.
    alpha : 'linear', float or array of floats
        If 'alpha' is a `float` or array of floats, this will be
        the transparency value. If it is 'linear' the transparency
        will decrease from '1' to '0'.

    """
    # Choose colormap
    if cmap is None:
        cmap = cm.viridis

    # Get the colormap colors
    my_cmap = cmap(np.arange(cmap.N))

    # Set alpha
    if alpha == "linear":
        my_cmap[:, -1] = np.arange(cmap.N) / cmap.N
    elif isinstance(alpha, float):
        my_cmap[:, -1] = np.ones(cmap.N) * alpha
    else:
        my_cmap[:, -1] = alpha

    # Create new colormap
    return ListedColormap(my_cmap)


# Disorder distributions
def standard_gaussian(label, salt, s1, s2=None):
    """Return a gaussian distribution."""
    return kwant.digest.gauss(str(hash((s1, s2))), salt=salt + label)


def uniform(label, salt, s1, s2=None):
    """Return a uniform distribution in the interval '[-0.5, 0.5)'."""
    return kwant.digest.uniform(str(hash((s1, s2))), salt=salt + label) - 0.5


def expectation_value(vectors, operator):
    """Braket 'operator' with 'vectors'.

    'vectors' must have t1he last axis as the dimension to be multiplied
    by 'operator'
    """
    dim = vectors.shape[-1]
    new_shape = vectors.shape[:-1]
    vecs = vectors.reshape(-1, dim)
    output = np.empty(new_shape, dtype=complex).flatten()
    for i, vec in enumerate(vecs):
        output[i] = np.dot(vec.conj(), operator.dot(vec))

    output = output.reshape(new_shape)
    return output


def local_expectation_value(vector, operator, norbs=2):
    """Trace over sites."""
    output = np.multiply(vector.conj(), operator.dot(vector.ravel()))
    return np.real_if_close(trace_orbs(output, norbs=norbs))



"""Main module containing iterables and generators of KPM vectors."""

from collections.abc import Iterable
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from math import ceil
from multiprocessing import Pool

import kwant
import numpy as np
from kwant.operator import _LocalOperator
from scipy.sparse import csr_matrix
from scipy.sparse import identity
from scipy.sparse.linalg import eigsh

from .common import identity_operator


DEFAULT_MOMENTS = 100


class IteratorKPM:
    r"""Iterator of KPM vectors.

    This iterator yields vectors as

    `T_n(H) | v >`

    for vectors `| v >` in `vectors`,
    for 'n' in '[0, max_moments]'.

    Notes
    -----
    Returns a sequence of expanded vectors of shape `(M, N)`.
    If the input is a vector then `M=1`.
    """

    def __init__(self, ham, vectors, max_moments=None, operator=None, num_threads=None):
        """Initialize the iterator.

        Parameters
        ----------
        ham : 2D array
            Hamiltonian, dense or sparse array with shape '(N, N)'.
        vectors : 1D or 2D array
            Vector of length 'N' or array of vectors with shape '(M, N)'.
        max_moments : int, optional
            Number of moments to stop the iteration. If not defined, the
            iterator has no end.
        operator : callable, optioinal
            Operator to act on the expanded vectors. The iterator will yield
            'operator(v_n)', where `v_n = T_n(H) | v >`. If omited, the
            identity operator is assumed.
        num_threads : int, optional
            Number of threads to use in matrix vector multiplications.
            If not provided, the maximum number of threads are used.

        """
        self.num_threads = num_threads
        self.alpha0 = np.array(np.atleast_2d(vectors).T, dtype=complex)
        self.len = max_moments if max_moments is not None else np.inf
        self.ham = ham
        self.n = 0
        self._operator = operator
        self.alpha_prev = None
        self.alpha = None

    def operator(self, vector):
        """Normalize operator."""
        if self._operator is None:
            return vector
        return self._operator(vector)

    def __len__(
        self,
    ):
        """Maximum number of moments and vectors to yield."""
        return self.len

    def add_moments(self, num_moments=0):
        """Increase the maximum number of moments returned.

        The iterator can now return `num_moments` new expanded vectors.
        Note that it will continue after the last vector, and not start from
        the beggining.
        """
        if num_moments >= 0 and num_moments == int(num_moments):
            self.len = self.len + num_moments
        else:
            raise ValueError("'num_moments' must be a positive integer.")

    def __iter__(
        self,
    ):
        """Return instance to iterate."""
        return self

    def __next__(
        self,
    ):
        """Yield the next vector."""
        if self.n >= self.len:
            raise StopIteration("Exhausted vectors.")
        if self.n == 0:
            self.alpha_prev = np.copy(self.alpha0)
            self.n = 1
            self.alpha = self.ham @ self.alpha_prev
            return self.operator(self.alpha_prev).T
        if self.n == 1:
            self.n = 2
            return self.operator(self.alpha).T
        if self.n >= 2:
            self.n = self.n + 1
            temp = self.alpha.copy()
            self.alpha = 2 * self.ham @ self.alpha - self.alpha_prev
            self.alpha_prev = temp

            return self.operator(self.alpha).T

    # TODO : add __getitem__ method to simplify calls in other modules,
    # specially `Correlator`

    def save(self):
        """Save internal state, except the 'ham'."""
        keys = ["n", "len", "alpha0", "alpha_prev", "alpha", "_operator"]
        return {k: self.__dict__[k] for k in keys}

    def load(self, d):
        """Load internal state."""
        self.__dict__.update(d)


def thread_update(spectrum, num_moments, executor=None, shutdown=True):
    """Update thread."""
    if executor is None:
        executor = ThreadPoolExecutor()

    all_moments = list(
        executor.map(
            spectrum._update_one_vector,
            np.arange(spectrum.num_vectors),
            [spectrum.num_moments + num_moments] * spectrum.num_vectors,
        )
    )
    executor.shutdown()

    for r in range(spectrum.num_vectors):
        spectrum._moments_list[r] = all_moments[r][0]
        spectrum._iterators[r].load(all_moments[r][1])

    spectrum.num_moments += num_moments

    spectrum._update_densities()


def process_update(spectrum, num_moments):
    """Update process."""
    with ProcessPoolExecutor() as executor:
        all_moments = list(
            executor.map(
                spectrum._update_one_vector,
                np.arange(spectrum.num_vectors),
                [spectrum.num_moments + num_moments] * spectrum.num_vectors,
            )
        )

    for r in range(spectrum.num_vectors):
        spectrum._moments_list[r] = all_moments[r][0]
        spectrum._iterators[r].load(all_moments[r][1])
    spectrum.num_moments = spectrum.num_moments + num_moments

    spectrum._update_densities()


def multiprocess_update(spectrum, num_moments):
    """Update multiprocess."""
    with Pool() as executor:
        all_moments = list(
            executor.starmap(
                spectrum._update_one_vector,
                [
                    (r, spectrum.num_moments + num_moments)
                    for r in range(spectrum.num_vectors)
                ],
            )
        )

    for r in range(spectrum.num_vectors):
        spectrum._moments_list[r] = all_moments[r][0]
        spectrum._iterators[r].load(all_moments[r][1])
    spectrum.num_moments += num_moments

    spectrum._update_densities()


def _normalize_num_moments(num_moments=None, energy_resolution=None, a=1):
    if (num_moments is not None) and (energy_resolution is not None):
        raise TypeError(
            "either 'num_moments' or 'energy_resolution' " "must be provided."
        )
    if energy_resolution is not None:
        if energy_resolution <= 0:
            raise ValueError("'energy resolution' must be positive")
        num_moments = ceil((1.6 * a) / energy_resolution)
    elif num_moments is None:
        num_moments = DEFAULT_MOMENTS

    if num_moments <= 0 or num_moments != int(num_moments):
        raise ValueError("'num_moments' must be a positive integer")

    return num_moments


class _BaseKPM:
    def __init__(
        self,
        hamiltonian,
        params=None,
        operator=None,
        num_vectors=10,
        num_moments=None,
        energy_resolution=None,
        vector_factory=None,
        bounds=None,
        eps=0.05,
        rng=None,
        kernel=None,
        mean=True,
        accumulate_vectors=True,
        num_threads=1,
    ):
        self.num_threads = num_threads

        # Normalize the format of 'ham'
        if isinstance(hamiltonian, kwant.system.System):
            hamiltonian = hamiltonian.hamiltonian_submatrix(params=params, sparse=True)
        try:
            hamiltonian = csr_matrix(hamiltonian)
        except Exception as err:
            raise ValueError(
                "'hamiltonian' is neither a matrix " "nor a Kwant system."
            ) from err

        # Normalize 'operator' to a common format.
        self.operator = _normalize_operator(operator, params)

        self.mean = mean
        rng0 = kwant._common.ensure_rng(rng)

        _v0 = None
        if bounds is None:
            # create this vector for reproducibility
            _v0 = np.exp(2j * np.pi * rng0.random_sample(hamiltonian.shape[0]))

        if eps <= 0:
            raise ValueError("'eps' must be positive")

        # Hamiltonian rescaled as in Eq. (24)
        self.hamiltonian, (self._a, self._b) = _rescale(
            hamiltonian, eps=eps, v0=_v0, bounds=bounds
        )
        self.bounds = (self._b - self._a, self._b + self._a)

        if vector_factory is None:
            self._vector_factory = kwant.kpm._VectorFactory(
                kwant.kpm.RandomVectors(hamiltonian, rng=rng),
                num_vectors=num_vectors,
                accumulate=accumulate_vectors,
            )
        else:
            if not isinstance(vector_factory, Iterable):
                raise TypeError("vector_factory must be iterable")
            try:
                len(vector_factory)
            except TypeError as err:
                if num_vectors is None:
                    raise ValueError(
                        "num_vectors must be provided if"
                        "vector_factory has no length."
                    ) from err
            self._vector_factory = kwant.kpm._VectorFactory(
                vector_factory, num_vectors=num_vectors, accumulate=accumulate_vectors
            )

        self._moments_list = []
        # sets self.num_vectors = 0
        self._iterators = []

        # set kernel before calling moments
        self.kernel = kernel if kernel is not None else kwant.kpm.jackson_kernel

        self.num_moments = _normalize_num_moments(
            num_moments, energy_resolution, self._a
        )

    @property
    def num_vectors(self):
        return len(self._iterators)

    def add_vectors(self, num_vectors=None):
        """Increase the number of vectors.

        If called with no arguments, or with 'num_vectors=None', then the
        vector factory is initialized and iterated to produce the vectors.

        Parameters
        ----------
        num_vectors: positive int, optional
            The number of vectors to add.

        """
        if num_vectors is None:
            num_vectors = self._vector_factory.num_vectors
        current_vectors = self.num_vectors
        total_vectors = current_vectors + num_vectors
        new_vectors_factory = total_vectors - self._vector_factory.num_vectors
        if new_vectors_factory > 0:
            self._vector_factory.add_vectors(new_vectors_factory)
        new_vectors = self._vector_factory.num_vectors - current_vectors

        if new_vectors <= 0 or new_vectors != int(new_vectors):
            raise ValueError("'num_vectors' must be a positive integer")

        self._iterators.extend([None] * new_vectors)

        for r in range(current_vectors, total_vectors):
            self._iterators[r] = IteratorKPM(
                self.hamiltonian, self._vector_factory[r], num_threads=self.num_threads
            )

    def save(self):
        """Save state."""
        exclude = ["hamiltonian", "_iterators"]
        d = {key: self.__dict__[key] for key in self.__dict__ if key not in exclude}

        iterators_data = [ite.save() for ite in self._iterators]
        d["iterators_data"] = iterators_data
        return d

    def load(self, d):
        """Load previous state."""
        if "iterators_data" in d:
            iterators = []
            iterators_data = d.pop("iterators_data")
            for data in iterators_data:
                ite = IteratorKPM(self.hamiltonian, data["alpha0"])
                ite.load(data)
                iterators.append(ite)
            d["_iterators"] = iterators

        self.__dict__.update(d)


class SpectralDensityIterator(_BaseKPM, kwant.kpm.SpectralDensity):
    """Inherit from ~`_BaseKPM` and ~`kwant.kpm.SpectralDensity`."""

    def __init__(self, *args, **kwargs):
        """Initialize Spectral Density Iterator.

        See ~`kwant.kpm.SpectralDensity`.
        """
        super().__init__(*args, **kwargs)

        self.add_vectors()
        self._update_moments_list(self.num_moments)
        self._update_densities()

    def _update_densities(self):
        moments = self._moments()
        self.densities, self._gammas = kwant.kpm._calc_fft_moments(moments)

    def add_moments(self, num_moments=None, *, energy_resolution=None):
        """Increase the number of Chebyshev moments.

        Parameters
        ----------
        num_moments: positive int
            The number of Chebyshev moments to add. Mutually
            exclusive with `energy_resolution`.
        energy_resolution: positive float, optional
            Features wider than this resolution are visible
            in the spectral density. Mutually exclusive with
            `num_moments`.

        """
        new_moments = _normalize_num_moments(num_moments, energy_resolution, self._a)

        self._update_moments_list(self.num_moments + new_moments)
        self.num_moments += new_moments

        # recalculate quantities derived from the moments
        self._update_densities()

    def add_vectors(self, num_vectors=None):
        """Increase the number of vectors.

        Parameters
        ----------
        num_vectors: positive int, optional
            The number of vectors to add.

        """
        current_vectors = self.num_vectors
        super().add_vectors(num_vectors=num_vectors)
        new_vectors = self.num_vectors - current_vectors

        self._moments_list.extend([[]] * new_vectors)

        for r in range(current_vectors, self.num_vectors):
            one_moment, _ = self._update_one_vector(r, self.num_moments)
            self._moments_list[r] = one_moment[:]

        self._update_densities()

    def _update_moments_list(self, n_moments):
        for r in range(self.num_vectors):
            one_moment, _ = self._update_one_vector(r, n_moments)
            self._moments_list[r] = one_moment[:]

    def _update_one_vector(self, r, n_moments):
        this_iterator = self._iterators[r]
        one_moment = [None] * n_moments
        one_moment[0 : this_iterator.n] = self._moments_list[r]

        for n in range(this_iterator.n, n_moments):
            alpha_next = next(this_iterator)
            one_moment[n] = np.dot(
                this_iterator.alpha0.conj().T, self.operator(alpha_next.T)
            ).squeeze()
        return one_moment, this_iterator.save()


class GeneralVectorExpansion(_BaseKPM):
    """Inherit from ~`_BaseKPM`."""

    def __init__(self, *args, coef_function=None, **kwargs):
        """Initialize general vector expansion."""
        super().__init__(*args, **kwargs)
        self._ab = (self._a, self._b)
        self.gs = self.kernel(np.ones(self.num_moments))

        if coef_function is None:
            raise NotImplementedError
        else:
            self.coef_function = coef_function

        self.add_vectors()

    def _initialize_iterators(
        self,
    ):
        for ite in self._iterators:
            ite.n = 0

    def __call__(self, e):
        """Return the general function spectrum for energy `e`."""
        iterator = self.vector_iterator(e)
        # vecs.shape = (num_vecs, dim)
        return sum(c[:, None, None] * vecs[None, :, :] for c, vecs in iterator)

    def vector_iterator(self, e):
        """Return vector iterator."""
        self._initialize_iterators()

        e = np.atleast_1d(e).flatten()

        # coef_array is finite and makes the zip iterator to stop
        coef_array = self.coef_function(self.num_moments, e, ab=self._ab)
        coef_array = self.gs[:, None] * coef_array

        for c in coef_array:
            vectors = []
            for r in range(self.num_vectors):
                vec = next(self._iterators[r])
                # vec.shape = (num_vecs_ite, dim)
                # where num_vecs_ite == 1
                vectors.append(vec.squeeze(0))
            yield c, np.array(vectors)

    def add_moments(self, num_moments=None, *, energy_resolution=None):
        """Increase the number of Chebyshev moments.

        Parameters
        ----------
        num_moments: positive int
            The number of Chebyshev moments to add. Mutually
            exclusive with `energy_resolution`.
        energy_resolution: positive float, optional
            Features wider than this resolution are visible
            in the spectral density. Mutually exclusive with
            `num_moments`.

        """
        new_moments = _normalize_num_moments(num_moments, energy_resolution, self._a)
        self.num_moments += new_moments
        for ite in self._iterators:
            ite.len += new_moments


def general_function(*args, **kwargs):
    """Build a general function operator using KPM.

    Returns a function that takes an energy or a list of energies, and returns
    the general function with that energy acting on `vectors`.

    Parameters
    ----------
    syst : kwant.System or ndarray
        Finalized kwant system or dense or sparse ndarray of the
        Hamiltonian with shape `(N, N)`.
    params : dict, optional
        Parameters for the kwant system.
    kpm_params : dict, optional
        Dictionary containing the parameters to pass to the `~kwant.kpm`
        module. 'num_vectors' will be overwritten to match the number
        of vectors, and 'operator' key will be deleted.
    precalculate_moments: bool, default False
        Whether to precalculate and store all the KPM moments of `vectors`.
        This is useful if the Green's function is evaluated at a large
        number of energies, but uses a large amount of memory.
        If False, the KPM expansion is performed every time the Green's
        function is called, which minimizes memory use.
    coef_function : callable, optional
        A function that takes a number of moments sequence of rescaled
        energies (to the interval '(-1, 1)'), and returns the coeficients
        of the expansion of the function.

    Returns
    -------
    f : callable
        Takes an energy or array of energies and returns the general function
        acting on the vectors, for those energies.

    """
    kpm_expansion = GeneralVectorExpansion(*args, **kwargs)
    init_vectors = np.array(kpm_expansion._vector_factory.saved_vectors)

    def f(e):
        """Return the function evaluated at 'e' applied to the vectors.

        The ndarray returned has initial dimensions '(N_E, M, N)',
        where 'N_E' is the number of energies passed, 'M' is the
        number of vectors, and 'N' is the dimension of the vectors.
        """
        vecs_in_energy = kpm_expansion(e)
        densities = np.einsum(
            "ijk,jk->ji",  # shape = (num_vectors, num_e)
            vecs_in_energy,
            init_vectors.conj(),
        )

        if kpm_expansion.mean:
            return np.mean(densities, axis=0)
        return densities

    return f


def concatenator(
    hamiltonian,
    params=None,
    operator_array=None,
    coef_array=None,
    return_expansion=False,
    first_transpose=False,
    ouroboros=False,
    logging=False,
    **kwargs
):
    """Expand a concatenation of functions and matrix operators.

    Take 'n' functions [f_i] and 'n+1' operators [O_j] and
    expand `<O_0 f_0 O_1 f_1... f_n 0_{n+1}>`.

    'O_j' is an operator acting on vectors, and [f_i] are functions
    defined by their coefficients on a KPM expansion.

    The instance created is a function. When called
    it returns the expectation value for a single energy.
    (maybe an array of energies too)

    Parameters
    ----------
    hamiltonian: ndarray or `kwant.system`
        Hamiltonian or system defining a Hamiltonian matrix.
    params: dict
        Parameters of the Hamiltonian or `kwant.system`.
    operator_array : iterable of operators
        The operators are normalized to an equivalent of a 'dot'
        function. If one of the elements in the iterable is 'None', then
        the identity operator is assumed. This identity is compatible
        with 'first_transpose'.
    coef_array: iterable of floats
        Array of coefficients for the expansion.
    return_expansion: bool, default to 'False'
        Return the expansion.
    first_transpose : bool, default to 'False'
        If there are more than one set of coefficients, then the first operator
        and set of coefficients are used to expand the vectors to the left.
        This means that the coefficients will be cojugated, and the that
        operator will be conjugate-transposed. This is achieved by different
        means, depending on the operator type. If the operator is a matrix, the
        conjugate-transpose of the matrix is taken. If it is a
        `~kwant.operator`, then this must be Hermitian. If it is a callable,
        then it must be a method from a numpy ndarray, or a scipy sparse
        matrix. In this case, the `__self__` attribute is attempted to
        retrieve, and the conjugate-transpose is taken.
        Mutually exclusive with 'ourorboros', in which case the saved expanded
        vectors to the right are the conjugate of the first expanded vectors to
        the left.
    ouroboros : bool, default to 'False'
        The first and last operators are equal, and the first and last
        functions defined by their KPM coefficients are also equal.
        Mutually exclusive with 'first_transpose'. Note that the operators and
        coefficients must nonetheless be passed, or at least place-holders.
    logging: bool, default to 'False'
        Use logging.
    Note
    ----
    The first operator in the operator array must be a matrix that can be
    transposed.

    """
    if first_transpose and ouroboros:
        raise ValueError("Only one of 'first_transpose' or 'ouroboros' can be True.")
    if len(coef_array) + 1 != len(operator_array):
        raise ValueError(
            "The length of the coefficients array must be the length of "
            "the operators minus one."
        )
    if (len(coef_array) == 1) and first_transpose:
        raise ValueError(
            "'first_transpose' cannot be true if only one operator " "is defined."
        )
    # Normalize 'operator_array' to functions that take and return
    # a vector.
    if logging:
        debug = ["First transpose is " + str(first_transpose)]
    operator_list = [
        _normalize_operator(operator_array[0], params, dagger=first_transpose)
    ]
    operator_list.extend(
        [_normalize_operator(op, params, dagger=False) for op in operator_array[1:]]
    )
    kwargs.pop("operator", None)  # no operators needed in kwargs

    # check that we accumulate the vectors from the factory
    if not kwargs.get("accumulate_vectors", True):
        raise ValueError("'accumulate_vectors' must be 'True'")
    kwargs["accumulate_vectors"] = True

    _temp = _BaseKPM(hamiltonian, params=params, **kwargs)
    _temp.add_vectors()
    # remove vector factory, since later is replaced with iterables of vectors
    kwargs.pop("vector_factory", None)

    vectors_left = np.array(_temp._vector_factory.saved_vectors)
    vectors_right = [operator_list[-1](vec) for vec in vectors_left]
    if logging:
        debug.append("vecs_right_0 = Op[-1](vecs)")

    if first_transpose:
        # first operator in the list is already dagger
        vectors_left = [operator_list[0](vec) for vec in vectors_left]
        if logging:
            debug.append("vecs_left_0 = vecs_right_0")

    # catch mean, and set it to False for the expansions
    mean = kwargs.get("mean", True)
    kwargs["mean"] = False

    if kwargs.get("bounds", None) is None:
        kwargs["bounds"] = _temp.bounds

    _psi = GeneralVectorExpansion(
        hamiltonian,
        params=params,
        coef_function=coef_array[-1],
        vector_factory=vectors_right,
        **kwargs
    )
    if logging:
        debug.append("vecs_right_1 = expand(coef[-1], vecs_right_0)")

    if len(coef_array) == 1:

        def f(e):
            e = np.atleast_1d(e)
            vecs_in_energy = _psi(e)
            # shape = (num_vectors, num_e)
            densities = np.einsum("ijk,jk->ji", vecs_in_energy, vectors_left.conj())

            if mean:
                densities = np.mean(densities, axis=0)

            if return_expansion:
                return densities, vectors_left, vecs_in_energy
            return densities

    elif len(coef_array) > 1:
        if first_transpose:
            _omega = GeneralVectorExpansion(
                hamiltonian,
                params=params,
                coef_function=_dagger(coef_array[0]),
                vector_factory=vectors_left,
                **kwargs
            )
            if logging:
                debug.append("vecs_left_1 = expand(coef[0].conj(), " "vecs_left_0)")
            remaining_coefs = coef_array[1:-1]
            remaining_operators = operator_list[1:-2]
        elif ouroboros:
            remaining_coefs = coef_array[1:-1]
            remaining_operators = operator_list[1:-2]
            if logging:
                debug.append("vecs_left_1 = vecs_right_1")
        else:
            remaining_coefs = coef_array[:-1]
            remaining_operators = operator_list[:-2]
            _vectors_left = vectors_left[np.newaxis].copy()

        if logging:
            # place-holders for next operations
            debug.extend([None, None, None])

        def f(e):
            e = np.atleast_1d(e)

            psi = _psi(e)

            if first_transpose:
                vectors_left = _omega(e)
            elif ouroboros:
                vectors_left = psi.copy()
            else:
                vectors_left = _vectors_left

            densities = []
            vecs_e_right = []
            for this_e, this_psi in zip(e, psi, strict=True):
                this_vecs_e_right = [operator_list[-2](vec) for vec in this_psi]
                if logging:
                    debug[-3] = "vecs_right_2 = Op[-2](vecs_right_1)"
                # flipping the order also flips which idx are included/excluded
                for coefs, op in zip(
                    remaining_coefs[::-1], remaining_operators[::-1], strict=True
                ):
                    _temp_expansion = GeneralVectorExpansion(
                        hamiltonian,
                        params=params,
                        coef_function=coefs,
                        vector_factory=this_vecs_e_right,
                        **kwargs
                    )
                    # shape=(num_e, num_vecs,dim)
                    _temp_vectors = _temp_expansion(this_e)[0]
                    if logging:
                        debug[-2] = "vecs_right_3 = expand(coef[-2], " "vecs_right_2)"
                    this_vecs_e_right = [op(vec) for vec in _temp_vectors]

                    if logging:
                        debug[-1] = "vecs_right_4 = Op[-3](vecs_right_3)"

                vecs_e_right.append(this_vecs_e_right)

            vecs_e_right = np.array(vecs_e_right)

            # shape=(num_vectors, num_e)
            densities = np.einsum("ijk,ijk->ji", vecs_e_right, vectors_left.conj())
            if mean:
                densities = np.mean(densities, axis=0)

            if logging:
                return densities, debug

            if return_expansion:
                return densities, vectors_left, vecs_e_right
            return densities

    return f


def _dagger(f):
    def dag(*args, **kwargs):
        return f(*args, **kwargs).conj()

    return dag


def _rescale(hamiltonian, eps, v0, bounds):
    """Rescale a Hamiltonian and return a LinearOperator.

    Parameters
    ----------
    hamiltonian : 2D array
        Hamiltonian of the system.
    eps : scalar
        Ensures that the bounds are strict.
    v0 : random vector, or None
        Used as the initial residual vector for the algorithm that
        finds the lowest and highest eigenvalues.
    bounds : tuple, or None
        Boundaries of the spectrum. If not provided the maximum and
        minimum eigenvalues are calculated.

    """
    # Relative tolerance to which to calculate eigenvalues.  Because after
    # rescaling we will add eps / 2 to the spectral bounds, we don't need
    # to know the bounds more accurately than eps / 2.
    tol = eps / 2

    if bounds:
        lmin, lmax = bounds
    else:
        lmax = float(
            eigsh(
                hamiltonian, k=1, which="LA", return_eigenvectors=False, tol=tol, v0=v0
            )
        )
        lmin = float(
            eigsh(
                hamiltonian, k=1, which="SA", return_eigenvectors=False, tol=tol, v0=v0
            )
        )

    a = np.abs(lmax - lmin) / (2.0 - eps)
    b = (lmax + lmin) / 2.0

    if lmax - lmin <= abs(lmax + lmin) * tol / 2:
        raise ValueError(
            "The Hamiltonian has a single eigenvalue, it is not possible to "
            "obtain a spectral density."
        )

    id = identity(hamiltonian.shape[0], format="csr")
    rescaled_ham = (hamiltonian - b * id) / a
    return rescaled_ham, (a, b)


def _normalize_operator(op, params, dagger=False):
    """Normalize 'op' to a function that takes and returns a vector."""
    if op is None:
        r_op = identity_operator
    elif isinstance(op, _LocalOperator):
        if dagger:
            if not op.check_hermiticity:
                raise ValueError(
                    "Cannot ensure that this operator is "
                    "Hermitian. Create this operator with "
                    "'check_hermiticity=True'"
                )
        op = op.bind(params=params)
        r_op = op.act
    elif callable(op):
        if dagger:
            try:
                r_op = op.__self__.T.conj().dot
            except Exception as err:
                raise ValueError(
                    "Cannot ensure that this operator is " "Hermitian."
                ) from err
        r_op = op
    elif hasattr(op, "dot"):
        op = csr_matrix(op)
        if dagger:
            r_op = op.T.conj().dot
        else:
            r_op = op.dot
    else:
        raise TypeError(
            "The operators must have a '.dot' " "attribute or must be callable."
        )
    return r_op
